%%
%% Morlet Wavelet Phase Method
%%
%
% This application implements the Morlet Wavelet Phase algorithm, developed
% primarily to measure frequency changes of Fabry Perot fringes in thin film
% optical reflectance spectra, to enable low limits of detection for optical
% thin film sensors. However, it can be used in any application where low
% noise measurement of frequency changes of an approximately sinusoidal
% signal are required. If you use this application, please cite the following
% paper which introduced the algorithm:
% 
% S. J. Ward, R. Layouni, S. Arshavsky-Graham, E. Segal, and S. M. Weiss,
% "Morlet Wavelet Filtering and Phase Analysis to Reduce the Limit of Detection
% for Thin Film Optical Biosensors,” ACS Sensors 6,  2967-2978 (2021).
%
% INSTRUCTIONS:
%Choose Wavelength Range of spectra to analyse (eg. 500nm - 800nm), enter
%into LowerLambdaBound and UpperLambdaBound. Enter index of spectra to use
%as reference (eg. 5 for the 5th spectra when alphanumerically ordered),
%enter ReferenceIndex. If there is a need to choose certain spectra files,
%enter common string of text in filename between ** in AllSpectraFiles (eg.
%if all spectra with "Exp" in the name should be analysed, *Exp*.txt). If
%all spectra are to be analysed AllSpectraFiles can be left as **.txt. To
%speed up the analysis, a minimum and maximum estimate for effective optical
%thickness (EOT), equal to 2nL (nm*RIU), can be provided by setting the
%variables LowerBoundEOTEstimate and UpperBoundEOTEstimate. This optional,
%but if used the film EOT must lie in this range. Wavelet parameters,
%FFT resolution and resolution in k-sace can also be optimized for a given
%measurement setup, but this is also optional. The result for each spectra
%is displayed, and stored in the Morlet Wavelet Phase (MWP) variable.
%
%Parameters:
%   *LowerLambdaBound: lower wavelength bound of the chosen spectral range
%    eg. 500nm (double)
%   *UpperLambdaBound: upper wavelength bound of the chosen spectral range
%    eg. 800nm (double)
%   *LowerBoundEOTEstimate: Lower bound for the estimated effective optical
%    thickness (2nL, where n is refractive index in RIU and L is thickness
%    in nm) of the thin film eg. 8000nm*RIU (double)
%   *UpperBoundEOTEstimate: Upper bound for the estimated effective optical
%    thickness (2nL, where n is refractive index in RIU and L is thickness
%    in nm) of the thin film eg. 20000nm*RIU (double)
%   *ReferenceIndex: Index of spectra file in alphanumerically sorted
%    spectra files in the current folder location chosen as the reference.
%    A baseline spectra in control solution. (int)
%Return:
%   *MWP: vector containing Morlet Wavelet Phase result for each of the N spectra files in
%   the current folder location (Ndarray)
%
%%
clear

%User selected parameters

LowerLambdaBound = 450;     %Lower bound of chosen spectral range with sufficient S/N (nm)
UpperLambdaBound = 900;       %Upper bound of chosen spectral range with sufficient S/N (nm)
LowerBoundEOTEstimate = 8000;     %OPTIONAL - Specify a lower bound to the expected 2nL value (um*RIU), to speed up search for peak 2nL
UpperBoundEOTEstimate = 20000;    %OPTIONAL - Specify a upper bound to the expected 2nL value (um*RIU), to speed up search for peak 2nL
AllSpectraFiles = dir('*P*.txt');              %Get all .txt files beginning with the string between the **. For example, '*Experiment*.txt' would fetch all .txt files beginning with "Experiment". If a single spectra is analyzed only the filename is needed, without any ** 
ReferenceFilename = 'P00000.txt';      %Specify reference spectra filename

%OPTIONAL - Parameters which can be optimized

kstep = 2E-7;          %Desired resolution in k-space used for interpolation of spectra
WaveletLength = 0.01;      %Length of wavelet in k-space
FFTZeroPaddingMultiplier = 2048;        %Multipler to increase zero padding for FFT

%Initializations

Desiredk = 1/UpperLambdaBound:kstep:1/LowerLambdaBound;        %Equally spaced k values for spectra interpolation
Waveletk  = -WaveletLength/2:kstep:WaveletLength/2;     %Wavelet k values
[~, SortedIndex]=sort({AllSpectraFiles.name});        %Index of alphanumerically sorted spectra files
AllSpectraFiles=AllSpectraFiles(SortedIndex);                       %Use index to alphanumerically sort spectra files
disp(['Reference Spectra Filename is ' ReferenceFilename]);        %Print reference spectra filename
Reference= dlmread(ReferenceFilename);     %Pick out reference spectra
RefLambda =Reference(:,1);RefR = Reference(:,2);      %Extract reference wavelength and reflectance for reference spectra
RefR = RefR - mean(RefR);       %Remove the mean from reference reflectance    
LongRefk = flip(1./RefLambda);     %Invert wavelength to obtain k values, and flip vector to get values in ascending order 
RefRFlip = flip(RefR);     %Flip reflectance vector to match up with flipped k-values
LambdaRange = find(RefLambda > LowerLambdaBound & RefLambda < UpperLambdaBound);      %Index corresponding to the chosen spectral range with sufficient S/N
RefRInterpolated = interp1(LongRefk, RefRFlip, Desiredk);        %Interpolated reflectance for reference spectra at every desired k value
RefRInterpolated = RefRInterpolated - mean(RefRInterpolated);       %remove mean from Interpolated reference spectra

%% Reference FFT

n = FFTZeroPaddingMultiplier*2^nextpow2(length(RefRInterpolated));      %Number of data points for FFT, determines resolution in frequency space, 
FFT = fft(RefRInterpolated, n);     %Carry out FFT of interpolated Reference spectra
frequency = (1/kstep)*(0:(n/2))/n;      %Points in frequency space
ReducedFrequencyRange = find(frequency>LowerBoundEOTEstimate & frequency<UpperBoundEOTEstimate);     %Find a reduced frequency range, between the lower and upper bounds that the 2nL value is guaranteed to be within
frequency = frequency(ReducedFrequencyRange);        %Cut down frequency range using estimated upper and lower frequency bouns
Power = abs(FFT/n);     %Normalize FFT output
Power = Power(1:n/2+1);     %Only considering positive frequencies
Power = Power(ReducedFrequencyRange);        %Only consider frequencies in the range estimated
[~,peakLocRef, FWHMRef] = findpeaks(Power,'SortStr','descend','WidthReference','halfheight');     %Find center frequency and FWHM of peaks (in order of descending height)
DominantFrequencyRef = frequency(peakLocRef(1));        %Use frequency of dominant peak

%% Reference Windowed FFT

RefRInterpolatedWindow = RefRInterpolated.*hann(length(RefRInterpolated),'periodic')';
FFTWindow = fft(RefRInterpolatedWindow, n);     %Carry out FFT of interpolated Reference spectra
PowerWindow = abs(FFTWindow/n);     %Normalize FFT output
PowerWindow = PowerWindow(1:n/2+1);     %Only considering positive frequencies
PowerWindow = PowerWindow(ReducedFrequencyRange);        %Only consider frequencies in the range estimated
[~,peakLocRefWindow, FWHMRefWindow] = findpeaks(PowerWindow,'SortStr','descend','WidthReference','halfheight');     %Find center frequency and FWHM of peaks (in order of descending height)

DominantFrequencyRefWindow = frequency(peakLocRefWindow(1));        %Use frequency of dominant peak

%% Reference Morlet Wavelet

WaveletExponentialRef = exp(1i*2*pi*DominantFrequencyRef.*Waveletk);     %Complex exponential with dominant frequency 
WaveletStdevRef =1/( 2*sqrt(2*log(2))*FWHMRef(1));      %Standard deviation of gaussian wavelet envelope, converting FFT FWHM to gaussian stdev
WGaussianEnvelopeRef  = exp((-Waveletk.^2)./(2*WaveletStdevRef^2));     %Define Gaussian wavelet envelope 
CMWRef = WaveletExponentialRef.*WGaussianEnvelopeRef;       %Reference Complex Morlet Wavelet
WaveletFilteredRef = conv(RefRInterpolated,CMWRef,'same');      %Reference Spectra filtered by convolution with the complex Morlet wavelet
WaveletFilteredRef = WaveletFilteredRef./(abs(WaveletFilteredRef));     %Remove envelope (complex amplitude) of filtered reference spectra (not necessary for LAMP method)

for i = 1:size(AllSpectraFiles,1)       %Iterate over all spectra files
    
    Data = dlmread(AllSpectraFiles(i).name);        %Load ith spectra file
    DataLambda =Data(:,1);DataR = Data(:,2);      %Extract reference wavelength and reflectance for ith spectra
    DataR = DataR - mean(DataR);      %Remove the mean from ith spectra reflectance
    LongDatak = flip(1./DataLambda);     %Invert wavelength to obtain k values, and flip vector to get values in ascending order 
    DataRFlip = flip(DataR);       %Flip reflectance vector to match up with flipped k-values
    
    %% Analyte FFT
    
    DataRInterpolated = interp1(LongDatak, DataRFlip, Desiredk);        %Interpolated reflectance for reference spectra at every desired k value
    DataRInterpolated = DataRInterpolated - mean(DataRInterpolated);        %Remove mean from Interpolated reference spectra
    n = FFTZeroPaddingMultiplier*2^nextpow2(length(RefRInterpolated));      %Number of data points for FFT, determines resolution in frequency space, 
    FFT = fft(DataRInterpolated, n);        %Carry out FFT of interpolated ith spectra
    frequency = (1/kstep)*(0:(n/2))/n;      %Points in frequency space
    ReducedFrequencyRange = find(frequency>LowerBoundEOTEstimate & frequency<UpperBoundEOTEstimate);      %Find a reduced frequency range, between the lower and upper bounds that the 2nL value is guaranteed to be within
    frequency = frequency(ReducedFrequencyRange);       %Cut down frequency range using estimated upper and lower frequency bouns
    Power = abs(FFT/n);     %Normalize FFT output
    Power = Power(1:n/2+1);     %Only consider positive frequencies
    Power = Power(ReducedFrequencyRange);       %Only consider frequencies in the range estimated
    [~,peakLoc, HWHM] = findpeaks(Power,'SortStr','descend','WidthReference','halfheight');       %Find center frequency and FWHM of peaks (in order of descending height)
    DominantFrequencyData = frequency(peakLoc(1));
    
    %% Analyte Windowed FFT

    DataRInterpolatedWindow = DataRInterpolated.*hann(length(DataRInterpolated),'periodic')';
    FFTWindow = fft(DataRInterpolatedWindow, n);     %Carry out FFT of interpolated Reference spectra
    PowerWindow = abs(FFTWindow/n);     %Normalize FFT output
    PowerWindow = PowerWindow(1:n/2+1);     %Only considering positive frequencies
    PowerWindow = PowerWindow(ReducedFrequencyRange);        %Only consider frequencies in the range estimated
    [~,peakLocWindow, HWHMWindow] = findpeaks(PowerWindow,'SortStr','descend','WidthReference','halfheight');     %Find center frequency and FWHM of peaks (in order of descending height)

    DominantFrequencyDataWindow = frequency(peakLocWindow(1))
    
    %% Analyte Complex Morlet Wavelet
    
    WaveletExponentialData = exp(1i*2*pi*DominantFrequencyData.*Waveletk);     %Complex exponential with dominant frequency 
    WaveletStdevData =1/( 2*sqrt(2*log(2))*HWHM(1));        %Standard deviation of gaussian wavelet envelope, converting FFT FWHM to gaussian stdev
    GaussianWEnvelopeData  = exp((-Waveletk.^2)./(2*WaveletStdevData^2));      %Define Gaussian wavelet envelope 
    CMWData = WaveletExponentialData.*GaussianWEnvelopeData;        %Complex Morlet Wavelet for ith spectra
    WeaveletFilteredR = conv(DataRInterpolated,CMWData,'same');        %ith spectra filtered by convolution with the complex Morlet wavelet
    WeaveletFilteredR = WeaveletFilteredR./(abs(WeaveletFilteredR));     %Remove envelope (complex amplitude) of filtered ith spectra (not necessary for LAMP method)

    %% Linear Average Morlet Phase
    
    %Extract unwrapped phase from complex Morlet wavelet filtered spectra
    
    RefAngle = unwrap(angle(WaveletFilteredRef));      %Unwrapped phase from filtered reference spectra
    DataAngle = unwrap(angle(WeaveletFilteredR));      %Unwrapped phase from filtered ith spectra
    
    %Ensure that the starting phase (RefAngle(1) and
    %DataAngle(1) are in the correct cycle relative to eachother
    
    RIFTSPrediction = (DominantFrequencyDataWindow - DominantFrequencyRefWindow)*0.0105 - 0.0655;   %Converting between optical thickness and the arbitrary Morlet Wavelet Phase units using predetermined calibration
    
    R_Difference = DataAngle - RefAngle;      %Phase difference at every wavenumber
    MWP(i) = mean(R_Difference);       %Morlet Wavelet Phase result, average difference in phase                       
    CycleDifference = round((MWP(i) - RIFTSPrediction)/(2*pi));    %Determining if the Morlet Wavelet Phase is more than a phase cycle different from optical thickness given by RIFTS
    MWP(i) = MWP(i) - CycleDifference*2*pi;     %Correct result if Morlet Wavelet Phase has moved into a 
    MWP(i) = (MWP(i) + 0.0655)/0.0105;      %Converting between arbitrary Morlet Wavelet Phase units and optical thickness using predetermined calibration
    
    disp(['Current spectra is ' num2str(AllSpectraFiles(i).name) ' MWP result is ' num2str(MWP(i))]);        %Print filename of ith spectra currently being analysed, and corresponding LAMP result
    plot(MWP, 'Color', [0, 0.4470, 0.7410])
    ylabel('Morlet Wavelet Phase \DeltaEffective Optical Thickness (nm)');
    xlabel('Spectrum Number');
    hold on;
    drawnow;
end